#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os.path

import semantic_version
import six

from muranopkgcheck import consts
from muranopkgcheck import error
from muranopkgcheck.i18n import _
from muranopkgcheck.validators import base
from muranopkgcheck import yaml_loader

error.register.E030(description='Not supported format')
error.register.E050(description='File is present in Manifest, '
                                'but not in filesystem')
error.register.E070(description='Tags should be a list')
error.register.E071(description='Type is invalid')
error.register.E072(description='UI is not a string')
error.register.E073(description='Invalid FullName')
error.register.E074(description='Logo is not a string')
error.register.W020(description='File is not present in Manifest, '
                                'but it is in filesystem')
error.register.W030(description='Not supported format version')
error.register.W073(description='There is no UI file')
error.register.W074(description='There is no Logo file')


class ManifestValidator(base.YamlValidator):
    def __init__(self, loaded_package):
        super(ManifestValidator, self).__init__(loaded_package,
                                                'manifest.yaml$')
        self.add_checker(self._valid_format, 'Format', False)
        self.add_checker(self._valid_string, 'Author', False)
        self.add_checker(self._valid_version, 'Version', False)
        self.add_checker(self._valid_fullname, 'FullName')
        self.add_checker(self._valid_string, 'Name', False)
        self.add_checker(self._valid_classes, 'Classes', False)
        self.add_checker(self._valid_tags, 'Tags', False)
        self.add_checker(self._valid_require, 'Require', False)
        self.add_checker(self._valid_type, 'Type')
        self.add_checker(self._valid_description, 'Description')
        self.add_checker(self._valid_ui, 'UI', False)
        self.add_checker(self._valid_logo, 'Logo', False)
        self.add_checker(self._valid_logo_ui_existance)

    def _valid_description(self, desc):
        if not isinstance(desc, six.string_types) and\
                not isinstance(desc, yaml_loader.YamlNull):
            yield error.report.E030('Value is not valid string "{0}"'
                                    .format(desc), desc)

    def _valid_format(self, value):
        format_ = str(value).split('/', 1)
        if len(format_) > 1:
            if format_[0] != 'MuranoPL':
                yield error.report.W030(_('Not supported format "{}"'
                                          '').format(value), value)
                return
        ver = format_[-1]
        if str(ver) not in ['1.0', '1.1', '1.2', '1.3', '1.4']:
            yield error.report.W030(_('Not supported format version "{}"'
                                      '').format(value), value)

    def _valid_fullname(self, fullname):
        if not self._check_fqn_name(fullname):
            yield error.report.E073(_('Invalid FullName "{}"')
                                    .format(fullname), fullname)

    def _valid_tags(self, value):
        if not isinstance(value, list):
            yield error.report.E070(_('Tags should be a list'), value)

    def _valid_require(self, value):
        if not isinstance(value, dict):
            yield error.report.E005(_('Require is not a dict type'), value)
            return
        for fqn, ver in six.iteritems(value):
            if not self._check_fqn_name(fqn):
                yield error.report.E005(_('Require key is not valid FQN "{}"'
                                          '').format(fqn), fqn)

    def _valid_type(self, value):
        if value not in ('Application', 'Library'):
            yield error.report.E071(_('Type is invalid "{}"').format(value),
                                    value)

    def _valid_version(self, version):
        try:
            semantic_version.Version.coerce(str(version))
        except ValueError:
            yield error.report.E071(_('Version format should be compatible '
                                      'with SemVer not "{}"'
                                      '').format(version), version)

    def _valid_logo_ui_existance(self, ast):
        if 'Logo' not in ast:
            yield self._valid_logo('logo.png')
        if 'UI' not in ast:
            yield self._valid_ui('ui.yaml')

    def _valid_ui(self, value):
        if isinstance(value, six.string_types):
            pkg_type = self._loaded_pkg.read(
                consts.MANIFEST_PATH).yaml()[0]['Type']
            if pkg_type == 'Library':
                return
            if not self._loaded_pkg.exists(os.path.join('UI', value)):
                yield error.report.W073(_('There is no UI file "{}"'
                                          '').format(value), value)
        else:
            yield error.report.E072(_('UI is not a string'), value)

    def _valid_logo(self, value):
        if isinstance(value, six.string_types):
            pkg_type = self._loaded_pkg.read(
                consts.MANIFEST_PATH).yaml()[0]['Type']
            if pkg_type == 'Library':
                return
            if not self._loaded_pkg.exists(value):
                yield error.report.W074(_('There is no Logo file "{}"'
                                          '').format(value), value)
        else:
            yield error.report.E074(_('Logo is not a string'), value)

    def _valid_classes(self, value):
        if not isinstance(value, dict):
            yield error.report.E074(_('Classes section should be a dict'),
                                    value)
            return

        files = set(value.values())
        existing_files = set(self._loaded_pkg.search_for('.*',
                                                         'Classes'))
        for fname in files - existing_files:
            yield error.report.E050(_('File "{}" is present in Manifest, '
                                      'but not in filesystem'
                                      '').format(fname), fname)
        for fname in existing_files - files:
            yield error.report.W020(_('File "{}" is not present in Manifest, '
                                      'but it is in filesystem'
                                      '').format(fname), fname)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy
import mock

from muranopkgcheck.tests import test_validator_helpers as helpers
from muranopkgcheck.validators import muranopl

MURANOPL_BASE = {
    'Name': 'Instance',
    'Namespaces': {
        '=': 'org.openstack.test',
        'res': 'io.murano.resources',
        'std': 'io.murano'},
    'Extends': 'res:LinuxMuranoInstance',
    'Properties': {
        'ports': {
            'Contract': '$.class(NeutronPort).notNull()',
            'Default': []}},
    'Methods': {
        'foo': {
            'Scope': 'Public',
            'Arguments': [{
                'arg1': {'Contract': '$.string()',
                         'Usage': 'Standard'}}],
            'Body': [
                {'Do': [
                    '$port.deploy()',
                    {'$template': {
                        'resources': {
                            '$.name': {
                                'properties': {
                                    'networks': [
                                        {'port': '$port.getRef()'}]}}}}},
                    {'$arg1':
                        '$arg1.mergeWith($template)'}],
                 'For': 'port',
                 'In': '$.ports'},
                {'$sth': 'new(res:Neutron)'},
                {'Return': '$arg1'}]}},
}


class MuranoPlTests(helpers.BaseValidatorTestClass):
    def setUp(self):
        super(MuranoPlTests, self).setUp()
        self.loaded_package = mock.Mock()
        self.loaded_package.format_version = '1.4'
        self.mpl_validator = muranopl.MuranoPLValidator(self.loaded_package)

    def test_import(self):
        self.g = self.mpl_validator._valid_import(['aaa.bbb', 'ccc.ddd',
                                                   'fff', 'w_ww'])

    def test_import_error(self):
        self.g = self.mpl_validator._valid_import(['aaa.bbb', 'ccc.ddd',
                                                   'fff', 'w_ww#'])
        self.assertIn('Wrong namespace or FNQ of extended class "w_ww#"',
                      next(self.g).message)

    def test_double_underscored_name(self):
        self.g = self.mpl_validator._valid_name('__Instance')
        self.assertIn('Invalid class name "__Instance"', next(self.g).message)

    def test_not_camel_case_name(self):
        self.g = self.mpl_validator._valid_name('notcamelcase')
        self.assertIn('Invalid class name "notcamelcase"',
                      next(self.g).message)

    def test_whitespace_in_name(self):
        name = 'white space'
        self.g = self.mpl_validator._valid_name(name)
        self.assertIn('Invalid class name "white space"',
                      next(self.g).message)

    def test_properties_list(self):
        self.g = self.mpl_validator._valid_properties([])
        self.assertIn('Properties should be a dict',
                      next(self.g).message)

    def test_properties_usage(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Usage'] = 'OutIn'
        self.g = self.mpl_validator._valid_properties(p_dict)
        self.assertIn('Not allowed usage "OutIn"',
                      next(self.g).message)

    def test_wrong_type_namespace(self):
        self.g = self.mpl_validator._valid_namespaces([1, 2, 3])
        self.assertIn('Wrong type of namespace',
                      next(self.g).message)

    def test_namespace_with_double_equal(self):
        self.g = self.mpl_validator._valid_namespaces({'==': 'std.io'})
        self.assertIn('Wrong name for namespace',
                      next(self.g).message)

    def test_namespace_with_wrong_fqn(self):
        self.g = self.mpl_validator._valid_namespaces({'=': 'io.murano@'})
        self.assertIn('Wrong namespace fqn "io.murano@"',
                      next(self.g).message)

    def test_wrong_method_scope(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        m_dict['foo']['Scope'] = 'Wrong'
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('Wrong Scope "Wrong"',
                      next(self.g).message)

    def test_dict_in_body(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        m_dict['foo']['Body'] = {'$a': 'b'}
        self.g = self.mpl_validator._valid_methods(m_dict)

    def test_error_in_method_scalar_body(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        m_dict['foo']['Body'] = '$.deploy('
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('"$.deploy(" is not valid yaql expression',
                      next(self.g).message)

    def test_method_body_is_return(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        m_dict['foo']['Body'] = {'Return': '3'}
        self.g = self.mpl_validator._valid_methods(m_dict)

    def test_error_in_method_for_loop_in(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        m_dict['foo']['Body'][0]['In'] =\
            '$.deploy('
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('"$.deploy(" is not valid yaql expression',
                      next(self.g).message)

    def test_error_in_method_for_loop_body(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        m_dict['foo']['Body'][0]['Do'][1] =\
            '$.deploy('
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('"$.deploy(" is not valid yaql expression',
                      next(self.g).message)

    def test_missing_contract_in_properties(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        del p_dict['ports']['Contract']
        self.g = self.mpl_validator._valid_properties(p_dict)
        self.assertIn('Missing Contract in property "ports"',
                      next(self.g).message)

    def test_contract_is_not_yaql(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Contract'] = '$.deploy('
        self.g = self.mpl_validator._valid_properties(p_dict)
        self.assertIn('Contract is not valid yaql "$.deploy("',
                      next(self.g).message)

    def test_contract_is_yaql_without_dollar(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Contract'] = 'deploy()'
        self.g = self.mpl_validator._valid_properties(p_dict)
        self.assertIn('Contract is not valid yaql "deploy()"',
                      next(self.g).message)

    def test_contract_is_yaql_without_dot(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Contract'] = '$deploy()'
        self.g = self.mpl_validator._valid_properties(p_dict)
        self.assertIn('Contract is not valid yaql "$deploy()"',
                      next(self.g).message)

    def test_contract_is_a_dict(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Contract'] = {
            '$.int()': '$.string()',
            '$.string()': ['$.ports()']
        }
        self.g = self.mpl_validator._valid_properties(p_dict)

    def test_contract_is_a_dict_with_two_levels(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Contract'] = {
            '$.int()': '$.string()',
            '$.string()': {'$.int()': '$.ports()', '$.int()': {
                '$': [], '$.int()': '$.string()'}}
        }
        self.g = self.mpl_validator._valid_properties(p_dict)

    def test_contract_is_a_dict_with_list(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Contract'] = {
            '$.int()': [],
            '$.int()': {}
        }
        self.g = self.mpl_validator._valid_properties(p_dict)

    def test_contract_list_with_numbers(self):
        self.g = self.mpl_validator._valid_contract(['$.string()', 1, 3])

    def test_contract_is_a_list_with_dict(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Contract'] = [{
            '$.string()': '$.string()',
            '$.int()': [{'$.int()': '$.string()', '$.int()': []}]}]
        self.g = self.mpl_validator._valid_properties(p_dict)

    def test_contract_is_a_list_two_elements(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Contract'] = ['$.string()', '$.int()']
        self.g = self.mpl_validator._valid_properties(p_dict)

    def test_contract_is_a_number(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Contract'] = 1
        self.g = self.mpl_validator._valid_properties(p_dict)
        self.assertIn('Contract is not valid yaql "1"',
                      next(self.g).message)

    def test_contract_list_with_min_length(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Contract'] = ['$.int()', 1]
        self.g = self.mpl_validator._valid_properties(p_dict)

    def test_contract_a_list_with_invalid_yaql(self):
        p_dict = deepcopy(MURANOPL_BASE['Properties'])
        p_dict['ports']['Contract'] = ['$.string(']
        self.g = self.mpl_validator._valid_properties(p_dict)
        self.assertIn('Contract is not valid yaql "$.string("',
                      next(self.g).message)

    def test_extends_is_not_a_valid_list(self):
        p_dict = deepcopy(MURANOPL_BASE['Extends'])
        p_dict = ['abc:def', 1]
        self.g = self.mpl_validator._valid_extends(p_dict)
        self.assertIn('Wrong type of Extends field',
                      next(self.g).message)

    def test_extends_fqn_wrong(self):
        p_dict = deepcopy(MURANOPL_BASE['Extends'])
        p_dict = ['abc:def.afdsa.das#']
        self.g = self.mpl_validator._valid_extends(p_dict)
        self.assertIn('Wrong FNQ of extended class "abc:def.afdsa.das#"',
                      next(self.g).message)

    def test_extends_is_not_valid(self):
        p_dict = deepcopy(MURANOPL_BASE['Extends'])
        p_dict = 4
        self.g = self.mpl_validator._valid_extends(p_dict)
        self.assertIn('Wrong type of Extends field',
                      next(self.g).message)

    def test_method_invalid_name(self):
        m_dict = {'foo#': {}}
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('Invalid name of method "foo#"',
                      next(self.g).message)

    def test_method_unknown_keyword(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        del m_dict['foo']['Body']
        m_dict['foo']['Body2'] = []
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('Unknown keyword "Body2"',
                      next(self.g).message)

    def test_body_number(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        m_dict['foo']['Body'] = 1
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('Body is not a list or scalar/yaql expression',
                      next(self.g).message)

    def test_dot_function(self):
        m_dict = {'.init': {}}
        self.g = self.mpl_validator._valid_methods(m_dict)

    def test_method_is_a_list(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        m_dict['foo'] = [1, 2]
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('Method is not a dict',
                      next(self.g).message)

    def test_method_scope_in_1_2(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        del m_dict['foo']['Arguments'][0]['arg1']['Usage']
        self.loaded_package.format_version = '1.2'
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('Scope is not supported version earlier than 1.3',
                      next(self.g).message)

    def test_method_usage_action_in_1_4(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        m_dict['foo']['Usage'] = 'Action'
        self.loaded_package.format_version = '1.4'
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('Usage "Action" is deprecated since 1.4',
                      next(self.g).message)

    def test_method_usage_action(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        m_dict['foo']['Usage'] = 'Action'
        self.loaded_package.format_version = '1.4'
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('Usage "Action" is deprecated since 1.4',
                      next(self.g).message)

    def test_method_wrong_usage_action(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        m_dict['foo']['Usage'] = 'Runtimed'
        self.loaded_package.format_version = '1.4'
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('Unsupported usage type "Runtimed"',
                      next(self.g).message)

    def test_method_wrong_usage_static_in_1_3(self):
        m_dict = deepcopy(MURANOPL_BASE['Methods'])
        del m_dict['foo']['Scope']
        del m_dict['foo']['Arguments'][0]['arg1']['Usage']
        m_dict['foo']['Usage'] = 'Static'
        self.loaded_package.format_version = '1.3'
        self.g = self.mpl_validator._valid_methods(m_dict)
        self.assertIn('Usage "Static" is available from 1.3',
                      next(self.g).message)

    def test_method_arguments_invalid(self):
        self.g = self.mpl_validator._valid_arguments('foobar')
        self.assertIn('Methods arguments should be a list or dict',
                      next(self.g).message)

    def test_method_arguments_are_dict(self):
        self.g = self.mpl_validator._valid_arguments({'a': {'Contract': '$'}})

    def test_method_arguments_are_dict_two_keys(self):
        self.g = self.mpl_validator._valid_arguments({'a': 'b', 'c': 'd'})
        self.assertIn('It is not safe to define methods arguments as a dict '
                      'with several keys', next(self.g).message)

    def test_method_arguments_element_is_two_key(self):
        arguments = [
            {'a': {'Contract': '$.string()'},
             'b': {'Contract': '$.string()'}}]
        self.g = self.mpl_validator._valid_arguments(arguments)
        self.assertIn('Methods single argument should be a one key dict',
                      next(self.g).message)

    def test_method_arguments_usage_1_3(self):
        self.loaded_package.format_version = '1.3'
        self.g = self.mpl_validator._valid_argument_usage('Standard')
        self.assertIn('Arguments usage is available since 1.4',
                      next(self.g).message)

    def test_wrong_method_arguments_usage(self):
        self.g = self.mpl_validator._valid_argument_usage('Standard1')
        self.assertIn('Usage is invalid value "Standard1"',
                      next(self.g).message)

    def test_method_arguments_invalid_name(self):
        self.g = self.mpl_validator._valid_arguments(
            [{'a#': {'Contract': '$.int()'}}])
        self.assertIn('Invalid name of argument "a#"',
                      next(self.g).message)



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import jinja2
import six

from muranopkgcheck import error

TEMPLATE = """
============  ================================================================
    CODE                           DESCRITION
============  ================================================================
{% for err in errors %}{{ err.code }}          {{err.description}}
{% endfor %}
============  ================================================================
"""


def main():
    from muranopkgcheck import validators
    print('Validators:')
    for v in validators.VALIDATORS:
        print(v.__name__)
    errors = sorted(six.itervalues(error.errors),
                    key=lambda item: item['code'])
    t = jinja2.Template(TEMPLATE)
    r = t.render(errors=errors)
    print(r)
    with open('doc/source/_errors_list.rst', 'w') as f:
        f.write(r)

if __name__ == '__main__':
    main()



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from muranopkgcheck.i18n import _LE

errors = dict()


class CheckError(Exception):

    def __init__(self, code, message, filename=None,
                 line=0, column=0, source=None):
        self.code = code
        self.message = message
        self.filename = filename
        self.line = line
        self.column = column
        self.source = source

    def to_dict(self):
        fields = ('code', 'message', 'filename', 'line', 'column', 'source')
        serialized = {}
        for f in fields:
            serialized[f] = self.__getattribute__(f)
        return serialized

    def is_warning(self):
        return self.code.split(':')[-1].startswith('W')

    def is_error(self):
        return self.code.split(':')[-1].startswith('E')

    def __repr__(self):
        return 'CheckError({0})'.format(self.message)


class Report(object):

    def __init__(self, errors, prefix=None):
        self.prefix = prefix
        self.errors = errors

    def __getattr__(self, code):
        code = ':'.join((self.prefix, code)) if self.prefix else code

        def _report(message, yaml_obj=None, filename=None):
            meta = getattr(yaml_obj, '__yaml_meta__', None)
            kwargs = {}
            if meta is not None:
                kwargs['line'] = meta.line + 1
                kwargs['column'] = meta.column + 1
                kwargs['source'] = meta.get_snippet()
                kwargs['filename'] = filename or meta.name
            return CheckError(code=code, message=message, **kwargs)
        if code not in self.errors:
            raise ValueError(_LE('Error {} was not registered').format(code))
        return _report


class Register(object):

    def __init__(self, errors, prefix=None):
        self.prefix = prefix
        self.errors = errors

    def __getattr__(self, code):
        code = ':'.join((self.prefix, code)) if self.prefix else code
        if code in self.errors:
            raise ValueError(_LE('Error {} is already registered')
                             .format(code))

        def _register(**kwargs):
            props = kwargs.copy()
            props['code'] = code
            self.errors[code] = props
        return _register

report = Report(errors)
register = Register(errors)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import itertools
import re

import six

from muranopkgcheck import error
from muranopkgcheck.i18n import _

FQN_REGEX = re.compile('^([a-zA-Z_$][\w$]*\.)*[a-zA-Z_$][\w$]*$')
NAME_REGEX = re.compile('^[A-Za-z_][\w]*$')

error.register.E005(description='YAML multi document is not allowed')
error.register.E020(description='Missing required key')
error.register.E021(description='Unknown keyword')
error.register.E040(description='Value should be string')
error.register.W010(description='Unknown keyword')


@six.add_metaclass(abc.ABCMeta)
class BaseValidator(object):

    def __init__(self, loaded_package, _filter='.*'):
        self._loaded_pkg = loaded_package
        self._filter = _filter

    @abc.abstractmethod
    def run(self):
        pass

    def _valid_string(self, value):
        if not isinstance(value, six.string_types):
            yield error.report.E040(_('Value is not a string "{}"'
                                      '').format(value), value)

    def _check_name(self, name):
        if isinstance(name, six.string_types) and NAME_REGEX.match(name):
            return True
        return False

    def _check_fqn_name(self, fqn):
        if isinstance(fqn, six.string_types) and FQN_REGEX.match(fqn):
            return True
        return False

    def _check_ns_fqn_name(self, ns_fqn):
        if isinstance(ns_fqn, six.string_types):
            if ':' in ns_fqn:
                ns, fqn = ns_fqn.split(':', 1)
                if NAME_REGEX.match(ns) and FQN_REGEX.match(fqn):
                    return True
            elif FQN_REGEX.match(ns_fqn):
                return True
        return False


class YamlValidator(BaseValidator):
    def __init__(self, loaded_package, _filter='.*', allows_multi=False):
        super(YamlValidator, self).__init__(loaded_package, _filter)
        self._checkers = {}
        self._allows_multi = allows_multi

    def add_checker(self, function, key=None, required=True):
        checkers = self._checkers.setdefault(key, {'checkers': [],
                                                   'required': False})
        checkers['checkers'].append(function)
        if key is None:
            checkers['required'] = False
        elif required:
            checkers['required'] = True

    def run(self):
        chain_of_suits = []
        for filename in self._loaded_pkg.search_for(self._filter):
            file_ = self._loaded_pkg.read(filename)
            chain_of_suits.append(self._run_single(file_))
        return itertools.chain(*chain_of_suits)

    def _run_single(self, file_):
        reports_chain = []

        def run_helper(name, checkers, data):
            for checker in checkers:
                result = checker(data)
                if result:
                    reports_chain.append(result)

        try:
            multi_documents = file_.yaml()
        except Exception as e:
            reports_chain.append([
                error.report.E002('Yaml Error: {0}'.format(e), e)])
        else:
            if multi_documents is None:
                multi_documents = [{}]

            if len(multi_documents) > 1 and not self._allows_multi:
                reports_chain.append([
                    error.report.E005(_('Multi document is not allowed in {}')
                                      .format(file_._path))])

            for ast in multi_documents:
                file_check = self._checkers.get(None)
                if file_check:
                    run_helper(None, file_check['checkers'], ast)
                for key, value in six.iteritems(ast):
                    checkers = self._checkers.get(key)
                    if checkers:
                        run_helper(key, checkers['checkers'], ast[key])
                    else:
                        reports_chain.append(self._unknown_keyword(key, value))
                missing = set(key for key, value in
                              six.iteritems(self._checkers)
                              if value['required']) - set(ast.keys())
                for m in missing:
                    reports_chain.append([
                        error.report.E020(_('Missing required key "{}"')
                                          .format(m), m)])

        return itertools.chain(*reports_chain)

    def _valid_keywords(self, present, known):
        unknown = set(present) - set(known)
        for u in unknown:
            yield error.report.E021(_('Unknown keyword "{}"').format(u), u)

    def _unknown_keyword(self, key, value):
        yield error.report.W010(_('Unknown keyword "{}"').format(key), key)

    def _null_checker(self, value):
        pass



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import types
import unittest


class BaseValidatorTestClass(unittest.TestCase):
    def setUp(self):
        self._g = []

    def tearDown(self):
        problems = [p for p in self.g]
        for p in problems:
            print('Left errors:', p)
        self.assertEqual(len(problems), 0)

    def _linear(self, error_chain):
        for e in error_chain:
            if isinstance(e, types.GeneratorType):
                for w in self._linear(e):
                    yield w
            else:
                yield e

    def get_g(self):
        return self._g

    def set_g(self, value):
        self._g = self._linear(value)

    g = property(get_g, set_g)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from muranopkgcheck.tests import test_validator_helpers as helpers
from muranopkgcheck.validators import ui


class UiValidatorTest(helpers.BaseValidatorTestClass):
    def setUp(self):
        super(UiValidatorTest, self).setUp()
        self.package = mock.Mock()
        self.ui_validator = ui.UiValidator(self.package)

    def test_forms(self):
        forms = [
            {'name1': {
                'fields': [{
                    'name': 'whatever',
                    'type': 'integer',
                    'label': 'sth',
                    'description': 'something'
                }]}}]
        self.g = self.ui_validator._valid_forms(forms)

    def test_forms_required_bool(self):
        forms = [
            {'name1': {
                'fields': [{
                    'name': 'whatever',
                    'type': 'integer',
                    'label': 'sth',
                    'description': 'something',
                    'required': 2,
                }]}}]
        self.g = self.ui_validator._valid_forms(forms)
        self.assertIn('Value of', next(self.g).message)

    def test_forms_hidden_bool(self):
        forms = [
            {'name1': {
                'fields': [{
                    'name': 'whatever',
                    'type': 'integer',
                    'label': 'sth',
                    'description': 'something',
                    'hidden': 2,
                }]}}]
        self.g = self.ui_validator._valid_forms(forms)
        self.assertIn('Value of', next(self.g).message)

    def test_version(self):
        self.g = self.ui_validator._valid_version('0.9')
        self.assertIn('Incorrect version of UI file "0.9"',
                      next(self.g).message)

    def test_aplication_list(self):
        self.g = self.ui_validator._valid_application([])
        self.assertIn('Application is not a dict',
                      next(self.g).message)

    def test_aplication_wrong_name(self):
        self.g = self.ui_validator._valid_application({
            '??': 'fasdf.dfas.dfd'})
        self.assertIn('Wrong name in UI file "??"',
                      next(self.g).message)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import zipfile

import six
import yaml
import yaml.error

from muranopkgcheck import consts
from muranopkgcheck import pkg_loader
from muranopkgcheck.tests import base


class FileWrapperTest(base.TestCase):

    @mock.patch('muranopkgcheck.pkg_loader.yaml')
    def test_file_wrapper(self, m_yaml):
        m_yaml.load_all.side_effect = yaml.load_all
        fake_pkg = mock.Mock()
        fake_pkg.open_file.side_effect = \
            lambda f: mock.mock_open(read_data='text')()
        f = pkg_loader.FileWrapper(fake_pkg, 'fake_path')
        self.assertEqual('text', f.raw())

        self.assertEqual(['text'], f.yaml())
        m_yaml.load_all.assert_called()

        self.assertEqual(['text'], f.yaml())
        m_yaml.load_all.reset_mock()
        m_yaml.load_all.assert_not_called()

        fake_pkg.open_file.side_effect = \
            lambda f: mock.mock_open(read_data='!@#$%')()
        f = pkg_loader.FileWrapper(fake_pkg, 'fake_path')
        self.assertEqual('!@#$%', f.raw())
        self.assertRaises(yaml.error.YAMLError, f.yaml)


class FakeLoader(pkg_loader.BaseLoader):

    @classmethod
    def _try_load(cls, path):
        return cls(path)

    def open_file(self, path, mode='r'):
        pass

    def exists(self, name):
        if name == consts.MANIFEST_PATH:
            return True

    def list_files(self, subdir=None):
        if subdir is None:
            return ['1.yaml', '2.sh', 'sub/3.yaml']
        else:
            return ['3.yaml']


class BaseLoaderTest(base.TestCase):

    @mock.patch.object(FakeLoader, 'read')
    @mock.patch.object(FakeLoader, '_try_load')
    @mock.patch.object(FakeLoader, 'try_set_format')
    def test_try_load(self, m_format, m_load, m_read):
        m_read.return_value.yaml.return_value = [{'FullName': 'fake'}]
        m_load.return_value = FakeLoader('fake')
        fake_pkg = FakeLoader.try_load('fake')
        self.assertEqual(m_load.return_value, fake_pkg)
        m_load.assert_called_once_with('fake')
        m_format.assert_called_once_with({'FullName': 'fake'})

        m_format.reset_mock()
        m_read.return_value.yaml.return_value = [{}]
        self.assertIsNone(FakeLoader.try_load('fake'))
        m_format.assert_not_called()

        m_format.reset_mock()
        m_read.return_value.yaml.side_effect = yaml.error.YAMLError()
        self.assertIsNone(FakeLoader.try_load('fake'))
        m_format.assert_not_called()

    @mock.patch.object(FakeLoader, '_try_load')
    def test_try_set_version(self, m_load):
        m_file_wrapper = mock.Mock()
        m_file = m_file_wrapper.return_value
        m_file.yaml.return_value = [{'Format': 'Fake/42', 'FullName': 'fake'}]
        with mock.patch('muranopkgcheck.pkg_loader.FileWrapper',
                        m_file_wrapper):
            m_load.return_value = FakeLoader('fake')
            loader = FakeLoader.try_load('fake')
            self.assertEqual('Fake', loader.format)
            self.assertEqual('42', loader.format_version)
            m_file.yaml.assert_called_once_with()

            m_load.return_value = FakeLoader('fake')
            m_file.yaml.return_value = [{'Format': '4.2', 'FullName': 'fake'}]
            loader = FakeLoader.try_load('fake')
            self.assertEqual(consts.DEFAULT_FORMAT, loader.format)
            self.assertEqual('4.2', loader.format_version)

            m_load.return_value = FakeLoader('fake')
            m_file.yaml.return_value = [{'FullName': 'fake'}]
            loader = FakeLoader.try_load('fake')
            self.assertEqual(consts.DEFAULT_FORMAT, loader.format)
            self.assertEqual(consts.DEFAULT_FORMAT_VERSION,
                             loader.format_version)

    def test_search_for(self):
        fake = FakeLoader('fake')
        self.assertEqual(['1.yaml', 'sub/3.yaml'],
                         list(fake.search_for('.*\.yaml$')))
        self.assertEqual(['3.yaml'],
                         list(fake.search_for('.*\.yaml$', subdir='sub')))

    def test_read(self):
        fake = FakeLoader('fake')
        m_file_wrapper = mock.Mock()
        m_file = m_file_wrapper.return_value
        with mock.patch('muranopkgcheck.pkg_loader.FileWrapper',
                        m_file_wrapper):
            loaded = fake.read('fake')
            self.assertEqual(m_file, loaded)
            # check that cache works
            loaded = fake.read('fake')
            self.assertEqual(m_file, loaded)


class DirectoryLoaderTest(base.TestCase):

    @mock.patch.object(pkg_loader.DirectoryLoader, 'read')
    @mock.patch.object(pkg_loader.DirectoryLoader, 'try_set_format')
    @mock.patch.object(pkg_loader.DirectoryLoader, 'exists')
    def _load_fake_pkg(self, m_exists, m_try_set_format, m_read):
        with mock.patch('muranopkgcheck.pkg_loader.os.path.isdir') as m_isdir:
            m_read.return_value.yaml.return_value = [{'FullName': 'fake'}]
            m_exists.return_value = True
            m_isdir.return_value = True
            loader = pkg_loader.DirectoryLoader.try_load('fake')
            m_try_set_format.assert_called_once_with({'FullName': 'fake'})
            return loader

    def test_try_load(self):
        # NOTE(sslypushenko) Using mock.patch here as decorator breaks pdb
        pkg = self._load_fake_pkg()
        self.assertEqual('fake', pkg.path)
        with mock.patch('muranopkgcheck.pkg_loader.os.path.isdir') as m_isdir:
            m_isdir.return_value = False
            pkg = pkg_loader.DirectoryLoader.try_load('fake')
            self.assertEqual(None, pkg)

    def test_list_files(self):
        # NOTE(sslypushenko) Using mock.patch here as decorator breaks pdb
        pkg = self._load_fake_pkg()
        with mock.patch('muranopkgcheck.pkg_loader.os.walk') as m_walk:
            m_walk.return_value = (item for item in [
                ('fake', ['subdir'], ['1', '2']),
                ('fake/subdir', [], ['3', '4']),
            ])
            self.assertEqual(['1', '2', 'subdir/3', 'subdir/4'],
                             pkg.list_files())
            m_walk.return_value = (item for item in [
                ('fake/subdir', [], ['3', '4']),
            ])
            self.assertEqual(['3', '4'],
                             pkg.list_files(subdir='subdir'))

    def test_exist(self):
        # NOTE(sslypushenko) Using mock.patch here as decorator breaks pdb
        pkg = self._load_fake_pkg()
        with mock.patch('muranopkgcheck.pkg_loader'
                        '.os.path.exists') as m_exists:
            m_exists.return_value = True
            self.assertTrue(pkg.exists('1.yaml'))
            m_exists.return_value = False
            self.assertFalse(pkg.exists('1.yaml'))


class ZipLoaderTest(base.TestCase):

    @mock.patch.object(pkg_loader.ZipLoader, 'read')
    @mock.patch.object(pkg_loader.ZipLoader, 'try_set_format')
    @mock.patch.object(pkg_loader.ZipLoader, 'exists')
    @mock.patch('muranopkgcheck.pkg_loader.zipfile')
    def _load_fake_pkg(self, m_zip, m_exists, m_try_set_format, m_read):
        m_zip_file = m_zip.ZipFile.return_value
        m_read.return_value.yaml.return_value = [{'FullName': 'fake'}]
        m_exists.return_value = True
        loader = pkg_loader.ZipLoader.try_load('fake')
        m_try_set_format.assert_called_once_with({'FullName': 'fake'})
        m_zip.ZipFile.assert_called_once_with('fake')
        return loader, m_zip_file

    def test_try_load(self):
        pkg, _ = self._load_fake_pkg()
        self.assertEqual('fake', pkg.path)
        with mock.patch('muranopkgcheck.pkg_loader.zipfile.ZipFile') as m_zip:
            m_zip.side_effect = zipfile.BadZipfile
            pkg = pkg_loader.ZipLoader.try_load('fake')
            self.assertEqual(None, pkg)

    @mock.patch.object(pkg_loader.ZipLoader, 'read')
    @mock.patch.object(pkg_loader.ZipLoader, 'try_set_format')
    @mock.patch.object(pkg_loader.ZipLoader, 'exists')
    @mock.patch('muranopkgcheck.pkg_loader.zipfile')
    def test_try_load_from_bytesio(self, m_zip, m_exists, m_try_set_format,
                                   m_read):
        m_read.return_value.yaml.return_value = [{'FullName': 'fake'}]
        m_exists.return_value = True
        m_content = str('fake') if six.PY2 else bytes('fake', 'latin-1')
        loader = pkg_loader.ZipLoader.try_load(six.BytesIO(m_content))
        self.assertIsNotNone(loader)
        m_try_set_format.assert_called_once_with({'FullName': 'fake'})
        m_zip.ZipFile.assert_called()

    def test_list_files(self):
        pkg, m_zip = self._load_fake_pkg()
        m_zip.namelist.return_value = ['1', '2', 'sub/', 'sub/3', 'sub/4']
        self.assertEqual(['1', '2', 'sub/3', 'sub/4'],
                         pkg.list_files())
        self.assertEqual(['3', '4'],
                         pkg.list_files(subdir='sub'))

    def test_exist(self):
        pkg, m_zip = self._load_fake_pkg()
        self.assertTrue(pkg.exists('1.yaml'))
        m_zip.getinfo.side_effect = KeyError
        self.assertFalse(pkg.exists('1.yaml'))
        m_zip.getinfo.side_effect = [KeyError, None]
        self.assertTrue(pkg.exists('sub'))
        m_zip.getinfo.side_effect = [KeyError, KeyError]
        self.assertFalse(pkg.exists('sub'))



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from muranopkgcheck import manager
from muranopkgcheck import plugin
from muranopkgcheck.tests import test_validator_helpers as helpers
from muranopkgcheck import validators


class FakePlugin(plugin.Plugin):

    def validators(self):
        pass

    def errors(self):
        pass


class PluginTest(helpers.BaseValidatorTestClass):

    def test_plugin(self):
        fake = FakePlugin()
        self.assertIsNotNone(fake)

    @mock.patch('muranopkgcheck.manager.pkg_loader')
    @mock.patch('muranopkgcheck.manager.stevedore')
    def test_load_plugins(self, m_stevedore, m_pkg_loader):
        fake_validator = mock.Mock()
        fake_plugin = mock.Mock()
        fake_plugin.obj.validators.return_value = [fake_validator]
        m_stevedore.ExtensionManager.return_value = [fake_plugin]
        fake_manager = manager.Manager('fake')
        m_pkg_loader.load_package.assert_called_once_with('fake', quiet=False)
        fake_manager.load_plugins()
        m_stevedore.ExtensionManager.assert_called()
        self.assertEqual(validators.VALIDATORS + [fake_validator],
                         fake_manager.validators)

        m_stevedore.ExtensionManager.reset_mock()
        fake_manager.load_plugins()
        m_stevedore.ExtensionManager.assert_not_called()



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import os
import sys

import six

from muranopkgcheck import log
from muranopkgcheck import manager

LOG = log.getLogger(__name__)


def parse_cli_args(args=None):

    usage_string = 'murano-pkg-checker [options] <path to package>'

    parser = argparse.ArgumentParser(
        description='murano-pkg-checker arguments',
        formatter_class=argparse.HelpFormatter,
        usage=usage_string
    )

    parser.add_argument('--select',
                        dest='select',
                        required=False,
                        type=str,
                        help='select errors and warnings (e.g. E001,W002)')

    parser.add_argument('--ignore',
                        dest='ignore',
                        required=False,
                        type=str,
                        help='skip errors and warnings (e.g. E042,W007)')

    parser.add_argument('--verbose', '-v',
                        dest='verbose',
                        default=0,
                        action='count',
                        help='Verbosity level. -v for ERROR. -vv for INFO')

    parser.add_argument('--debug',
                        dest='debug',
                        action='store_true',
                        help='Set logging level to DEBUG')

    parser.add_argument('--discover',
                        dest='discover',
                        action='store_true',
                        help='Run discovery packages')

    parser.add_argument('path',
                        type=str,
                        help='Path to package or catalog')

    return parser.parse_args(args=args)


def setup_logs(args):
    if args.verbose == 0:
        log.setup(level=log.CRITICAL)
    elif args.verbose == 1:
        log.setup(level=log.ERROR)
    else:
        log.setup(level=log.INFO)
    if args.debug:
        log.setup(level=log.DEBUG)


def run(args, pkg_path=None, quiet_load=False):
    m = manager.Manager(pkg_path or args.path, quiet_load=quiet_load)
    m.load_plugins()
    if args.select:
        select = args.select.split(',')
    else:
        select = None
    if args.ignore:
        ignore = args.ignore.split(',')
    else:
        ignore = None
    errors = m.validate(select=select, ignore=ignore)
    fmt = manager.PlainTextFormatter()
    return fmt.format(errors)


def discover(args):
    errors = []
    for dirpath, dirnames, filenames in os.walk(args.path):
        items = dirnames
        for item in items:
            if item.startswith('.'):
                continue
            try:
                path = os.path.join(dirpath, item)
                pkg_errors = run(args, path, quiet_load=True)
                LOG.info("Package {} discovered".format(path))
                if pkg_errors:
                    errors.append("Errors in package {}\n{}\n"
                                  "".format(path, pkg_errors))
            except ValueError:
                pass
    return '\n'.join(errors)


def main():
    args = parse_cli_args()
    setup_logs(args)
    try:
        if args.discover:
            errors = discover(args)
        else:
            errors = run(args)
    except ValueError as e:
        LOG.error(six.text_type(e))
        print(six.text_type(e))
        return 2
    except Exception as e:
        LOG.critical(six.text_type(e), exc_info=sys.exc_info())
        return 3
    if errors:
        print(errors)
        return 1
    else:
        print('No errors found!')

if __name__ == '__main__':
    sys.exit(main())



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import six

from muranopkgcheck.checkers import code_structure
from muranopkgcheck.checkers import yaql_checker
from muranopkgcheck import error
from muranopkgcheck.i18n import _
from muranopkgcheck.validators import base


SUPPORTED_FORMATS = frozenset(['1.0', '1.1', '1.2', '1.3', '1.4'])
METHOD_KEYWORDS = frozenset(['Body', 'Arguments', 'Usage', 'Scope', 'Meta'])
METHOD_ARGUMENTS_KEYWORDS = frozenset(['Contract', 'Usage', 'Meta', 'Default'])
PROPERTIES_KEYWORDS = frozenset(['Contract', 'Usage', 'Default', 'Meta'])
PROPERTIES_USAGE_VALUES = frozenset(['In', 'Out', 'InOut', 'Const', 'Static',
                                     'Runtime', 'Config'])
CLASSNAME_REGEX = re.compile('^[A-Za-z_]\w*$')
METHOD_NAME_REGEX = re.compile('^[A-Za-z_\.][\w]*$')

error.register.E011(description='Invalid class name')
error.register.E025(description='Wrong namespace or FNQ of extended class')
error.register.E026(description='Properties should be a dict')
error.register.E042(description='Not allowed usage')
error.register.E044(description='Wrong type of namespace')
error.register.E045(description='Body is not a list or scalar/yaql expression')
error.register.E046(description='Method is not a dict or list')
error.register.E047(description='Missing Contract in property')
error.register.E048(description='It is not safe to define methods arguments '
                                'as a dict with several keys')
error.register.E052(description='Arguments usage is available since 1.4')
error.register.E053(description='Usage is invalid value ')
error.register.E054(description='Invalid name of method "{}"')
error.register.E060(description='Wrong namespace fqn')
error.register.W045(description='Unsupported usage type')
error.register.W011(description='Invalid class name')
error.register.W048(description='Contract is not valid yaql')


class MuranoPLValidator(base.YamlValidator):
    def __init__(self, loaded_package):
        super(MuranoPLValidator, self).__init__(loaded_package,
                                                'Classes/.*\.yaml$',
                                                True)
        self.yaql_checker = yaql_checker.YaqlChecker()
        self.code_structure = code_structure.CheckCodeStructure()

        self.add_checker(self._null_checker, 'Meta', False)
        self.add_checker(self._valid_name, 'Name', False)
        self.add_checker(self._valid_extends, 'Extends', False)
        self.add_checker(self._valid_methods, 'Methods', False)
        self.add_checker(self._valid_import, 'Import', False)
        self.add_checker(self._valid_namespaces, 'Namespaces', False)
        self.add_checker(self._valid_properties, 'Properties', False)

    def _valid_import(self, import_, can_be_list=True):
        if can_be_list and isinstance(import_, list):
            for imp in import_:
                yield self._valid_import(imp, False)
        elif not self._check_ns_fqn_name(import_):
            yield error.report.E025(_('Wrong namespace or FNQ of extended '
                                      'class "{0}"').format(import_), import_)

    def _valid_name(self, value):
        if value.startswith('__') or \
           not CLASSNAME_REGEX.match(value):
            yield error.report.E011(_('Invalid class name "{}"').format(value),
                                    value)
        elif not (value != value.lower() and value != value.upper()):
            yield error.report.W011(_('Invalid class name "{}"').format(value),
                                    value)

    def _valid_extends(self, value, can_be_list=True):
        if can_be_list and isinstance(value, list):
            for cls in value:
                yield self._valid_extends(cls, False)
        elif isinstance(value, six.string_types):
            if not self._check_ns_fqn_name(value):
                yield error.report.E025(_('Wrong FNQ of extended class "{}"'
                                          '').format(value), value)
        else:
            yield error.report.E025("Wrong type of Extends field", value)

    def _valid_contract(self, contract):
        if isinstance(contract, list):
            if len(contract) > 1:
                if len(contract) < 3:
                    if isinstance(contract[1], int):
                        return
                elif len(contract) < 4:
                    if isinstance(contract[1], int) and \
                            isinstance(contract[2], int):
                        return
                for con in contract:
                    yield self._valid_contract(con)
            elif len(contract) == 1:
                yield self._valid_contract(contract[0])
        elif isinstance(contract, dict):
            if not contract:
                return
            for c_key, c_value in six.iteritems(contract):
                yield self._valid_string(c_key)
                yield self._valid_contract(c_value)
        elif isinstance(contract, six.string_types):
            if not self.yaql_checker(contract) or \
                    not contract.startswith('$.') and contract != '$':
                yield error.report.W048(_('Contract is not valid yaql "{}"'
                                          '').format(contract), contract)
        else:
            yield error.report.W048(_('Contract is not valid yaql "{}"'
                                      '').format(contract), contract)

    def _valid_properties(self, properties):
        if not isinstance(properties, dict):
            yield error.report.E026(_('Properties should be a dict'),
                                    properties)
            return
        for property_name, property_data in six.iteritems(properties):
            usage = property_data.get('Usage')
            if usage:
                if usage not in PROPERTIES_USAGE_VALUES:
                    yield error.report.E042(_('Not allowed usage "{}"'
                                              '').format(usage), usage)
            contract = property_data.get('Contract')
            if contract is not None:
                yield self._valid_contract(contract)
            else:
                yield error.report.E047(_('Missing Contract in property "{}"')
                                        .format(property_name), property_name)
            yield self._valid_keywords(property_data.keys(),
                                       PROPERTIES_KEYWORDS)

    def _valid_namespaces(self, value):
        if not isinstance(value, dict):
            yield error.report.E044(_('Wrong type of namespace'), value)
            return

        for name, fqn in six.iteritems(value):
            if not self._check_fqn_name(fqn):
                yield error.report.E060(_('Wrong namespace fqn '
                                          '"{}"').format(fqn), fqn)
            if not self._check_name(name) and name != '=':
                yield error.report.E060(_('Wrong name for namespace '
                                          '"{}"').format(fqn), fqn)

    def _valid_methods(self, value):
        for method_name, method_data in six.iteritems(value):
            if not isinstance(method_data, dict):
                if method_data:
                    yield error.report.E046(_('Method is not a dict'),
                                            method_name)
                return

            if not METHOD_NAME_REGEX.match(method_name):
                yield error.report.E054(_('Invalid name of method "{}"')
                                        .format(method_name), method_name)
            scope = method_data.get('Scope')
            if scope:
                yield self._valid_scope(scope)
            usage = method_data.get('Usage')
            if usage:
                yield self._valid_method_usage(usage)
            arguments = method_data.get('Arguments')
            if arguments:
                yield self._valid_arguments(arguments)
            body = method_data.get('Body')
            if body:
                yield self._valid_body(body)
            yield self._valid_keywords(method_data.keys(), METHOD_KEYWORDS)

    def _valid_body(self, body):
        if not isinstance(body, (list, six.string_types, dict)):
            yield error.report.E045(_('Body is not a list or scalar/yaql '
                                    'expression'), body)
        else:
            yield self.code_structure.codeblock(body)

    def _valid_scope(self, scope):
        if self._loaded_pkg.format_version >= '1.4':
            if scope is not None and scope not in ('Public', 'Session'):
                yield error.report.E044(_('Wrong Scope "{}"').format(scope),
                                        scope)
        else:
            yield error.report.E044(_('Scope is not supported version '
                                    'earlier than 1.3"'), scope)

    def _valid_method_usage(self, usage):
        if usage == 'Action':
            if self._loaded_pkg.format_version >= '1.4':
                yield error.report.W045(_('Usage "{}" is deprecated since 1.4'
                                          '').format(usage), usage)
        elif usage in frozenset(['Static', 'Extension']):
            if self._loaded_pkg.format_version <= '1.3':
                yield error.report.W045(_('Usage "{}" is available from 1.3')
                                        .format(usage), usage)
        elif usage != 'Runtime':
            yield error.report.W045(_('Unsupported usage type "{}" ')
                                    .format(usage), usage)

    def _valid_arguments(self, arguments):
        if isinstance(arguments, dict) and len(arguments) > 1:
            yield error.report.E048(_('It is not safe to define methods '
                                      'arguments as a dict with several keys'),
                                    arguments)
            return
        elif not isinstance(arguments, (list, dict)):
            yield error.report.E046(_('Methods arguments should be a list or '
                                      'dict'), arguments)
            return
        if isinstance(arguments, dict):
            arguments = [arguments]
        for argument in arguments:
            if not isinstance(argument, dict) or len(argument) != 1:
                yield error.report.E046(_('Methods single argument should be '
                                          'a one key dict'), argument)
            else:
                name = next(six.iterkeys(argument))
                if not self._check_name(name):
                    yield error.report.E054(_('Invalid name of argument "{}"')
                                            .format(name), name)
                val = next(six.itervalues(argument))
                contract = val.get('Contract')
                if contract:
                    yield self._valid_contract(contract)
                usage = val.get('Usage')
                if usage:
                    yield self._valid_argument_usage(usage)
                yield self._valid_keywords(val, METHOD_ARGUMENTS_KEYWORDS)

    def _valid_argument_usage(self, usage):
        if self._loaded_pkg.format_version < '1.4':
            yield error.report.E052(_('Arguments usage is available '
                                      'since 1.4'), usage)
        if usage not in frozenset(['Standard', 'VarArgs', 'KwArgs']):
            yield error.report.E053(_('Usage is invalid value "{}"')
                                    .format(usage), usage)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from muranopkgcheck.checkers import code_structure
from muranopkgcheck.tests import test_validator_helpers as helpers


class CodeStructureTest(helpers.BaseValidatorTestClass):
    def setUp(self):
        super(CodeStructureTest, self).setUp()
        self._checker = code_structure.CheckCodeStructure()

    def test_simple(self):
        SIMPLE_BODY = '$.deploy()'
        self.g = self._checker.codeblock(SIMPLE_BODY)

    def test_double_assigment(self):
        SIMPLE_BODY = [{
            '$a': '$.deploy()',
            '$b': '$.string()'}]
        self.g = self._checker.codeblock(SIMPLE_BODY)
        self.assertIn('Wrong code structure/assigment. Probably a typo',
                      next(self.g).message)

    def test_multiline(self):
        MULTILINE_BODY = [
            '$.deploy()',
            {'$res': 'new(YaqlStuff)'},
            '$.call($res)',
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)

    def test_bad_assigment(self):
        MULTILINE_BODY = [
            '$.deploy()',
            {1: 'new(YaqlStuff)'},
            '$.call($res)',
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)
        self.assertIn('"1" is not valid variable name', next(self.g).message)

    def test_bad_assigment_with_double_dollar(self):
        MULTILINE_BODY = [
            '$.deploy()',
            {'$$': 'new(YaqlStuff)'},
            '$.call($res)',
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)
        self.assertIn('"$$" is not valid variable name', next(self.g).message)

    def test_bad_assigment_case2(self):
        MULTILINE_BODY = [
            '$.deploy()',
            {'res': 'new(YaqlStuff)'},
            '$.call($res)',
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)
        p = next(self.g)
        self.assertIn('"res" is not valid variable name', p.message)

    def test_if(self):
        MULTILINE_BODY = [
            {'If': '$.deploy()',
             'Then': [
                 '$.w()',
                 {'$abc': '$a'}]}
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)

    def test_while_missing_do(self):
        MULTILINE_BODY = [
            {'While': '$.deploy()'}
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)
        p = next(self.g)
        self.assertIn('Missing keyword "Do" for "While" code structure',
                      p.message)

    def test_while_unknown_does(self):
        MULTILINE_BODY = [
            {'While': '$.deploy()',
             'Does': ['$.a()', '$.b()']}
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)
        p1 = next(self.g)
        p2 = next(self.g)
        six.assertCountEqual(self, [
            'Unknown keyword "Does" in "While"',
            'Missing keyword "Do" for "While" code structure'],
            [p1.message, p2.message])

    def test_empty_return(self):
        MULTILINE_BODY = [
            {'Return': ''}
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)

    def test_switch(self):
        MULTILINE_BODY = [
            {'Switch': {
                '$.black()': '$.single()',
                '$.blue()': [
                    '$.b()',
                    {'$w': 3}]},
             'Default': '$.a()'}
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)

    def test_error_under_while_in_if(self):
        MULTILINE_BODY = [
            {'If': '1',
             'Then': {'While': '$.deploy()',
                      'Do': [
                          {'www': '$.a()'},
                          '$.b()']}}
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)
        self.assertIn('"www" is not valid variable name',
                      next(self.g).message)

    def test_minimal_try_block(self):
        MULTILINE_BODY = [
            {'Try': [
                '$port.deploy()'],
             'Catch': {}}]
        self.g = self._checker.codeblock(MULTILINE_BODY)

    def test_try_not_string(self):
        MULTILINE_BODY = [
            {'Try': ['$port.deploy()'],
             'Catch': {
                'With': 'exceptionName',
                'As': 213,
                'Do': ['$.string()']}}
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)
        self.assertIn('Value of "213" should be a string',
                      next(self.g).message)

    def test_yaql_accept_bool(self):
        self.g = self._checker.yaql(True)

    def test_not_empty(self):
        MULTILINE_BODY = [
            '$.deploy()',
            {'$d': 'new(YaqlStuff)'},
            '$.call($res)',
            {'Break': 'a'},
        ]
        self.g = self._checker.codeblock(MULTILINE_BODY)
        self.assertIn('Statement should be empty, not a "a"',
                      next(self.g).message)



#    Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import os

import oslotest.base
import six
import testscenarios
import yaml

from muranopkgcheck import consts
from muranopkgcheck import manager
from muranopkgcheck import pkg_loader


class DictLoader(pkg_loader.BaseLoader):

    @classmethod
    def _try_load(cls, pkg):
        if consts.MANIFEST_PATH in pkg:
            return cls(pkg)
        return None

    def __init__(self, pkg):
        super(DictLoader, self).__init__('')
        self.pkg = pkg

    def open_file(self, path, mode='r'):
        if self.pkg[path]['format'] == 'raw':
            sio = io.StringIO(six.text_type(self.pkg[path]['content']))
            setattr(sio, 'name', path)
        elif self.pkg[path]['format'] == 'yaml':
            content = yaml.safe_dump(self.pkg[path]['content'])
            sio = io.StringIO(six.text_type(content))
            setattr(sio, 'name', path)
        else:
            raise ValueError('Unknown type of content')
        return sio

    def list_files(self, subdir=None):
        files = self.pkg.keys()
        if subdir is None:
            return files
        subdir_len = len(subdir)
        return [file_[subdir_len:].lstrip('/') for file_ in files
                if file_.startswith(subdir)]

    def exists(self, name):
        return name in self.pkg


class DictFormatter(manager.Formatter):

    def format(self, error):
        return sorted([{'code': e.code, 'msg': e.message} for e in error],
                      key=lambda item: item['code'])


def load_cases():
    cases_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                              'cases')
    cases_files = [os.path.join(cases_path, f)for f in os.listdir(cases_path)
                   if os.path.isfile(os.path.join(cases_path, f))]
    cases = []
    for cases_file in cases_files:
        with open(cases_file) as f:
            cases.extend(list(yaml.load_all(f)))
    return cases

cases = load_cases()


class TestCase(testscenarios.WithScenarios, oslotest.base.BaseTestCase):

    """Test case base class for all unit tests."""

    scenarios = cases

    def test_foo(self):
        m = manager.Manager(self.pkg, loader=DictLoader)
        errors = m.validate()
        fmt = DictFormatter()
        self.assertEqual(self.expected, fmt.format(errors))



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from muranopkgcheck import error
from muranopkgcheck.i18n import _
from muranopkgcheck.validators import base

KNOWN_FILES_DIR = frozenset(['manifest.yaml', 'UI', 'LICENSE', 'Classes',
                             'images.lst', 'README.rst'])
REQUIRED_FILES_DIR = frozenset(['manifest.yaml', 'LICENSE'])

error.register.W120(description='Unknown file in the package')
error.register.W121(description='Missing file in the package')


class PackageValidator(base.BaseValidator):
    def __init__(self, loaded_package):
        super(PackageValidator, self).__init__(loaded_package, '')

    def run(self):
        yield self._known_directories()

    def _known_directories(self):
        files = set(self._loaded_pkg.search_for('^[^/]+$'))
        try:
            logo_file = next(self._loaded_pkg.search_for('^manifest.yaml$'))\
                .yaml()[0]['Logo']
        except Exception:
            logo_file = 'logo.png'
        for file_ in files - KNOWN_FILES_DIR - {logo_file}:
            yield error.report.W120(_('Unknown "{}" in the package')
                                    .format(file_), file_)
        for file_ in REQUIRED_FILES_DIR - files:
            yield error.report.W121(_('Missing "{}" in the package')
                                    .format(file_), file_)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import unittest

from muranopkgcheck.validators import base


class YamlValidatorTest(unittest.TestCase):
    def setUp(self):
        self.pkg = mock.Mock()
        self.pkg.search_for.return_value = ['sth']
        self.fmock = mock.Mock()
        self.document = mock.Mock()
        self.pkg.read.return_value = self.fmock
        self.v = base.YamlValidator(self.pkg, '***')

    def test_checker_with_ast(self):
        c = mock.Mock()
        c.return_value = 'ok'
        self.fmock.yaml.return_value = [{}]
        self.v.add_checker(c)
        self.v.run()
        c.assert_called_once_with({})
        self.pkg.search_for.assert_called_once_with('***')

    def test_run_single_with_key_checker(self):
        c = mock.Mock()
        c.return_value = 'ok'
        self.fmock.yaml.return_value = [{'key': 'whatever'}]
        self.v.add_checker(c, 'key')
        self.v.run()
        c.assert_called_once_with('whatever')
        self.pkg.search_for.assert_called_once_with('***')

    def test_two_keys_unknown_key(self):
        c = mock.Mock()
        c.return_value = None
        self.fmock.yaml.return_value = [{'key': 'whatever',
                                         'unknown': ''}]
        self.v.add_checker(c, 'key')
        errors = self.v.run()
        c.assert_called_once_with('whatever')
        self.pkg.search_for.assert_called_once_with('***')
        self.assertIn('Unknown keyword "unknown"', next(errors).message)

    def test_missing_required_key(self):
        c = mock.Mock()
        self.fmock.yaml.return_value = [{}]
        self.v.add_checker(c, 'key')
        errors = self.v.run()
        self.pkg.search_for.assert_called_once_with('***')
        self.assertIn('Missing required key "key"', next(errors).message)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaql

ITERATORS_LIMIT = 100
EXPRESSION_MEMORY_QUOTA = 512 * 1024

ENGINE_10_OPTIONS = {
    'yaql.limitIterators': ITERATORS_LIMIT,
    'yaql.memoryQuota': EXPRESSION_MEMORY_QUOTA,
    'yaql.convertSetsToLists': True,
    'yaql.convertTuplesToLists': True,
    'yaql.iterableDicts': True
}

ENGINE_12_OPTIONS = {
    'yaql.limitIterators': ITERATORS_LIMIT,
    'yaql.memoryQuota': EXPRESSION_MEMORY_QUOTA,
    'yaql.convertSetsToLists': True,
    'yaql.convertTuplesToLists': True
}


def _add_operators(engine_factory):
    engine_factory.insert_operator(
        '>', True, 'is',
        yaql.factory.OperatorType.BINARY_LEFT_ASSOCIATIVE, False)
    engine_factory.insert_operator(
        None, True, ':',
        yaql.factory.OperatorType.BINARY_LEFT_ASSOCIATIVE, True)
    engine_factory.insert_operator(
        ':', True, ':',
        yaql.factory.OperatorType.PREFIX_UNARY, False)
    engine_factory.operators.insert(0, ())


def _create_engine():
    engine_factory = yaql.factory.YaqlFactory()
    _add_operators(engine_factory=engine_factory)
    options = ENGINE_12_OPTIONS
    return engine_factory.create(options=options)


class YaqlChecker(object):
    def __init__(self):
        self._engine = _create_engine()

    def __call__(self, data):
        try:
            self._engine(data)
        except yaql.utils.exceptions.YaqlParsingException:
            return False
        except TypeError:
            return False
        return True



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock

from muranopkgcheck.tests import test_validator_helpers as helpers
from muranopkgcheck.validators import package


class PackageValidatorTests(helpers.BaseValidatorTestClass):
    def setUp(self):
        super(PackageValidatorTests, self).setUp()
        self.loaded_package = mock.Mock()
        self.pv = package.PackageValidator(self.loaded_package)

    def test_known_files(self):
        self.loaded_package.search_for.return_value = [
            'manifest.yaml', 'LICENSE', 'logo']
        self.g = self.pv.run()
        self.assertIn('Unknown "logo" in the package',
                      next(self.g).message)

    def test_known_files_missing_req(self):
        self.loaded_package.search_for.return_value = [
            'manifest.yaml', 'logo.png']
        self.g = self.pv.run()
        self.assertIn('Missing "LICENSE" in the package',
                      next(self.g).message)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from muranopkgcheck import log
from muranopkgcheck.tests import base


class LogTest(base.TestCase):

    @mock.patch('muranopkgcheck.log.logging')
    def test_get_logger(self, m_log):
        fake_logger = log.getLogger('fake')
        m_log.getLogger.assert_called_once_with('fake')
        self.assertEqual(fake_logger, log._loggers['fake'])
        self.assertEqual(fake_logger, m_log.getLogger.return_value)
        m_log.reset_mock()
        fake_logger = log.getLogger('fake')
        m_log.assert_not_called()
        self.assertEqual(fake_logger, log._loggers['fake'])

    @mock.patch('muranopkgcheck.log.logging')
    def test_setup(self, m_log):
        fake_handler = m_log.StreamHandler.return_value
        fake_formatter = m_log.Formatter.return_value
        fake_logger = mock.Mock()
        fake_old_handler = mock.Mock()
        fake_logger.handlers = [fake_old_handler]
        log._loggers['fake'] = fake_logger

        log.setup(log_format='fake_format', level=42)

        m_log.StreamHandler.assert_called_once_with()
        fake_handler.setFormatter.assert_called_once_with(fake_formatter)
        fake_logger.setLevel.assert_called_once_with(42)
        fake_logger.removeHandler.assert_called_once_with(fake_old_handler)
        fake_logger.addHandler.assert_called_once_with(fake_handler)

    def test_setup_ext_logging(self):
        fake_logging = mock.Mock()
        log.setup(external_logging=fake_logging)
        self.assertEqual(log._logging, fake_logging)
        log.getLogger('fake_name')
        fake_logging.getLogger.assert_called_once_with('fake_name')



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from muranopkgcheck import error
from muranopkgcheck.tests import base


class LogTest(base.TestCase):

    @mock.patch('muranopkgcheck.error.CheckError')
    def test_errors(self, m_error):
        errors = dict()
        register = error.Register(errors, prefix='PRE')
        report = error.Report(errors, prefix='PRE')
        register.F042(description='Fake error')
        self.assertRaises(ValueError, register.__getattr__,
                          code='F042')
        self.assertEqual({'PRE:F042': {'code': 'PRE:F042',
                                       'description': 'Fake error'}},
                         errors)
        yaml_obj = mock.MagicMock(__yaml_meta__=mock.Mock(line=1,
                                                          column=1))
        type(yaml_obj.__yaml_meta__).name = mock.PropertyMock(
            return_value='fake.yaml')
        yaml_obj.__yaml_meta__.get_snippet.return_value = 'fake_code'
        report.F042('It is an error!', yaml_obj)
        m_error.assert_called_once_with(
            code='PRE:F042',
            column=2,
            line=2,
            filename='fake.yaml',
            message='It is an error!',
            source='fake_code')
        self.assertRaises(ValueError, report.__getattr__,
                          code='FAKE')



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import six

CRITICAL = logging.CRITICAL
ERROR = logging.ERROR
WARNING = logging.WARNING
INFO = logging.INFO
DEBUG = logging.DEBUG

LOG_FORMAT = "%(name)s:%(lineno)d %(levelname)s %(message)s"
DEFAULT_LEVEL = logging.DEBUG
_loggers = {}
_logging = None


def setup(external_logging=None, log_format=LOG_FORMAT, level=DEFAULT_LEVEL):

    if external_logging:
        global _logging
        _logging = external_logging
        return

    console_log_handler = logging.StreamHandler()
    console_log_handler.setFormatter(logging.Formatter(log_format))
    global _loggers
    for logger in six.itervalues(_loggers):
        logger.setLevel(level)
        for h in logger.handlers:
            logger.removeHandler(h)
        logger.addHandler(console_log_handler)


def getLogger(name):
    global _logging, _loggers
    if _logging:
        return _logging.getLogger(name)
    if name not in _loggers:
        _loggers[name] = logging.getLogger(name)
    return _loggers[name]



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from muranopkgcheck import error
from muranopkgcheck import manager
from muranopkgcheck.tests import base


class PlainTextFormatterTest(base.TestCase):

    def test_format(self):

        class FakeYamlMeta(object):

            def __init__(self):
                self.line = 0
                self.column = 0
                self.name = 'fake'

            def get_snippet(self):
                return 'fake'

        class FakeYamlNode(str):

            def __init__(self, value):
                super(FakeYamlNode, self).__init__()
                self.value = value
                self.__yaml_meta__ = FakeYamlMeta()

        formatter = manager.PlainTextFormatter()
        fake_yaml_node = FakeYamlNode('Fake!!!')
        errors = [error.report.E042('Fake!!!', fake_yaml_node)]
        self.assertEqual('fake:1:1: E042 Fake!!!', formatter.format(errors))


class ManagerTest(base.TestCase):

    def _assert_errors(self, expected, actual):
        expected_errors = list(expected)
        for e in actual:
            if e in expected_errors:
                expected_errors.remove(e)
            else:
                self.fail('Unexpected error {}'.format(e))
        self.assertEqual([], expected_errors, 'Expected errors left')

    @mock.patch('muranopkgcheck.manager.pkg_loader')
    @mock.patch('muranopkgcheck.manager.error.report')
    def test_validate(self, m_error, m_pkg_loader):
        fake_error = error.CheckError('W007', 'Fake')
        m_error.W007.return_value = fake_error
        fake_E000_error = error.CheckError('E000', 'Fake')
        m_error.E000.return_value = fake_E000_error

        def error_generator():
            yield fake_error

        def broken_checker():
            a = 0
            if 1 / a:
                yield fake_error

        def broken_method_checker(self):
            a = 0
            if 1 / a:
                yield fake_error

        mgr = manager.Manager('fake')
        m_pkg_loader.load_package.assert_called_once_with('fake', quiet=False)
        MockValidator = mock.Mock()
        m_validator = MockValidator.return_value

        def prepare_errors():
            return iter((fake_error, error_generator(), broken_checker(),
                         broken_method_checker(m_validator)))

        m_validator.run.return_value = prepare_errors()
        errors = mgr.validate(validators=[MockValidator])
        self._assert_errors(
            [fake_E000_error, fake_E000_error, fake_error, fake_error],
            errors)

        m_validator.run.return_value = prepare_errors()
        errors = mgr.validate(validators=[MockValidator], only_errors=True)
        self._assert_errors([fake_E000_error, fake_E000_error], errors)

        m_validator.run.return_value = prepare_errors()
        errors = mgr.validate(validators=[MockValidator], select=['W007'])
        self._assert_errors([fake_error, fake_error], errors)

        m_validator.run.return_value = prepare_errors()
        errors = mgr.validate(validators=[MockValidator], ignore=['E000'])
        self._assert_errors([fake_error, fake_error], errors)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import abc
import io
import os
import re
import sys
import zipfile

import six
import yaml

from muranopkgcheck import consts
from muranopkgcheck.i18n import _
from muranopkgcheck import log
from muranopkgcheck import yaml_loader

LOG = log.getLogger(__name__)


class FileWrapper(object):

    def __init__(self, pkg, path):
        self._path = path
        with pkg.open_file(path) as file_:
            self._raw = file_.read()
            self._name = file_.name
        self._yaml = None
        self._pkg = pkg

    def raw(self):
        return self._raw

    def yaml(self):
        if self._yaml is None:
            sio = io.StringIO(six.text_type(self.raw()))
            setattr(sio, 'name', self._name)
            self._yaml = list(yaml.load_all(sio,
                                            yaml_loader.YamlLoader))
        return self._yaml


@six.add_metaclass(abc.ABCMeta)
class BaseLoader(object):
    def __init__(self, path):
        self.path = path
        self._cached_files = dict()
        self.format = consts.DEFAULT_FORMAT
        self.format_version = consts.DEFAULT_FORMAT_VERSION

    @classmethod
    @abc.abstractmethod
    def _try_load(cls, path):
        pass    # pragma: no cover

    @classmethod
    def try_load(cls, path):
        loader = cls._try_load(path)
        if loader is not None and loader.exists(consts.MANIFEST_PATH):
            try:
                manifest = loader.read(consts.MANIFEST_PATH).yaml()[0]
                if 'FullName' not in manifest:
                    LOG.warning('Package does not look like Murano package',
                                exc_info=sys.exc_info())
                    return
                loader.try_set_format(manifest)
            except yaml.YAMLError:
                LOG.warning('Unable to parse Manifest yaml',
                            exc_info=sys.exc_info())
                return
            return loader

    @abc.abstractmethod
    def list_files(self, subdir=None):
        pass    # pragma: no cover

    @abc.abstractmethod
    def open_file(self, path, mode='r'):
        pass    # pragma: no cover

    @abc.abstractmethod
    def exists(self, name):
        pass    # pragma: no cover

    def search_for(self, regex='.*', subdir=None):
        r = re.compile(regex)
        return (f for f in self.list_files(subdir) if r.match(f))

    def read(self, path):
        if path in self._cached_files:
            return self._cached_files[path]
        self._cached_files[path] = FileWrapper(self, path)
        return self._cached_files[path]

    def try_set_format(self, manifest):
        if manifest and 'Format' in manifest:
            if '/' in six.text_type(manifest['Format']):
                fmt, version = manifest['Format'].split('/', 1)
                self.format = fmt
                self.format_version = version
            else:
                self.format_version = six.text_type(manifest['Format'])


class DirectoryLoader(BaseLoader):

    @classmethod
    def _try_load(cls, path):
        if os.path.isdir(path):
            return cls(path)
        return None

    def open_file(self, path, mode='r'):
        return open(os.path.join(self.path, path), mode)

    def list_files(self, subdir=None):
        path = self.path
        if subdir is not None:
            path = os.path.join(path, subdir)

        files = []
        for dirpath, dirnames, filenames in os.walk(path):
            files.extend(
                os.path.relpath(
                    os.path.join(dirpath, filename), self.path)
                for filename in filenames)
        if subdir is None:
            return files
        subdir_len = len(subdir)
        return [file_[subdir_len:].lstrip('/') for file_ in files]

    def exists(self, name):
        return os.path.exists(os.path.join(self.path, name))


class ZipLoader(BaseLoader):

    def __init__(self, path):
        super(ZipLoader, self).__init__(path)
        if hasattr(self.path, 'read'):
            self._zipfile = zipfile.ZipFile(six.BytesIO(self.path.read()))
        else:
            self._zipfile = zipfile.ZipFile(self.path)

    @classmethod
    def _try_load(cls, path):
        try:
            return cls(path)
        except (IOError, zipfile.BadZipfile):
            return None

    def open_file(self, name, mode='r'):
        return self._zipfile.open(name, mode)

    def list_files(self, subdir=None):
        files = [file_ for file_ in self._zipfile.namelist()
                 if not file_.endswith('/')]
        if subdir is None:
            return files
        subdir_len = len(subdir)
        return [file_[subdir_len:].strip('/') for file_ in files
                if file_.startswith(subdir)]

    def exists(self, name):
        try:
            self._zipfile.getinfo(name)
            return True
        except KeyError:
            pass

        if not name.endswith('/'):
            try:
                self._zipfile.getinfo(name + '/')
                return True
            except KeyError:
                pass
        return False


PACKAGE_LOADERS = [DirectoryLoader, ZipLoader]


def load_package(path, quiet=False):
    for loader_cls in PACKAGE_LOADERS:
        loader = loader_cls.try_load(path)
        if loader is not None:
            return loader
        else:
            if not quiet:
                LOG.debug("{} failed to load '{}'"
                          "".format(loader_cls.__name__, path))
    else:
        raise ValueError(_('Can not load package: "{}"').format(path))



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from muranopkgcheck import error
from muranopkgcheck.i18n import _
from muranopkgcheck.validators import base

UI_VERSION = frozenset(('1.0', '1', '2', '2.0', '2.1', '2.2', '2.3'))
FIELDS_TYPE = frozenset(('string', 'boolean', 'text', 'integer', 'password',
                         'clusterip', 'floatingip', 'domain', 'databaselist',
                         'table', 'flavor', 'keypair', 'image', 'azone',
                         'psqlDatabase', 'network', 'choice'))
BOOL_FIELDS = frozenset(('required', 'hidden'))
STR_FIELDS = frozenset(('name', 'label', 'description',
                        'descriptionTitle', 'regexpValidator', 'helpText'))
INT_FIELDS = frozenset(('minLength', 'maxLength', 'minValue', 'maxValue'))

error.register.E081(description='Value should be boolean')
error.register.E083(description='Wrong name in UI file')
error.register.E084(description='Application is not a dict')
error.register.E100(description='Not valid FQN or known type')
error.register.W082(description='Incorrect version of UI file')
error.register.W100(description='Not known type. Probably typo')


class UiValidator(base.YamlValidator):
    def __init__(self, loaded_package):
        super(UiValidator, self).__init__(loaded_package, 'UI/.*\.yaml$')
        self.add_checker(self._valid_forms, 'Forms', False)
        self.add_checker(self._null_checker, 'Templates', False)
        self.add_checker(self._valid_application, 'Application', False)
        self.add_checker(self._valid_version, 'Version', False)

    def _valid_application(self, application):
        if not isinstance(application, dict):
            yield error.report.E084(_('Application is not a dict'),
                                    application)
            return
        for name, value in six.iteritems(application):
            if not self._check_name(name):
                if name != '?':
                    yield error.report.E083(_('Wrong name in UI file "{}"')
                                            .format(name), name)

    def _valid_version(self, version):
        if str(version) not in UI_VERSION:
            yield error.report.W082(_('Incorrect version of UI file "{}"')
                                    .format(version), version)

    def _valid_forms(self, forms):
        for named_form in forms:
            for name, form in six.iteritems(named_form):
                yield self._valid_form(form['fields'])
                yield self._valid_keywords(form.keys(),
                                           ('fields', 'validators'))

    def _valid_form(self, form):
        for named_params in form:
            for key, value in six.iteritems(named_params):
                if key in STR_FIELDS:
                    if not isinstance(value, six.string_types):
                        yield error.report.E040(_('Value of {} should be '
                                                  'string not "{}"')
                                                .format(key, value), key)
                elif key in BOOL_FIELDS:
                    if not isinstance(value, bool):
                        yield error.report.E081(_('Value of {} should be '
                                                  'boolean not "{}"')
                                                .format(key, value), key)
                elif key in INT_FIELDS:
                    if not isinstance(value, int):
                        yield error.report.E082(_('Value of {} should be '
                                                  'int not "{}"')
                                                .format(key, value), key)
                elif key == 'type':
                    yield self._valid_field_type(value)

    def _valid_field_type(self, fqn, can_be_list=True):
        if isinstance(fqn, list):
            for elem in fqn:
                yield self._valid_field_type(elem, False)
        elif self._check_fqn_name(fqn):
            if '.' not in fqn and fqn not in FIELDS_TYPE:
                yield error.report.W100('"{0}" is not known type. '
                                        'Probably a typo'.format(fqn), fqn)
        else:
            yield error.report.E100('"{0}" is not valid FQN or known type'
                                    .format(fqn), fqn)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import itertools
import pprint
import sys
import types

import six
import stevedore

from muranopkgcheck import error
from muranopkgcheck.i18n import _
from muranopkgcheck.i18n import _LE
from muranopkgcheck import log
from muranopkgcheck import pkg_loader
from muranopkgcheck.validators import VALIDATORS

LOG = log.getLogger(__name__)

error.register.E000(description='Check failed')


@six.add_metaclass(abc.ABCMeta)
class Formatter(object):

    @abc.abstractmethod
    def format(self, error):
        pass    # pragma: no cover


class PlainTextFormatter(Formatter):

    def format(self, errors):
        lines = []
        for e in errors:
            if e.filename:
                lines.append('{filename}:{line}:{column}: {code} {message}'
                             ''.format(**e.to_dict()))
            else:
                lines.append('{code} {message}'
                             ''.format(**e.to_dict()))

        return '\n'.join(lines)


class Manager(object):

    def __init__(self, pkg_path, quiet_load=False, loader=None):
        if loader:
            self.pkg = loader(pkg_path)
        else:
            self.pkg = pkg_loader.load_package(pkg_path, quiet=quiet_load)
        self.validators = list(VALIDATORS)
        self.plugins = None

    def _to_list(self, error_chain, select=None, ignore=None):
        errors = []
        while True:
            try:
                e = next(error_chain)
            except StopIteration:
                break
            except Exception:
                exc_info = sys.exc_info()
                tb = exc_info[2]
                while tb.tb_next:
                    tb = tb.tb_next
                validator_class = tb.tb_frame.f_locals.get('self')
                check_name = tb.tb_frame.f_code.co_name
                check_locals = tb.tb_frame.f_locals.copy()
                check_locals.pop('self', None)
                if validator_class:
                    msg = (_('Checker {} from {} failed!'
                             '').format(check_name,
                                        validator_class.__class__.__name__))
                else:
                    msg = (_('Checker {} failed!'
                             '').format(check_name))
                LOG.error('{} {}\n{}'.format(msg,
                                             _('Checker locals:'),
                                             pprint.pformat(check_locals)),
                          exc_info=exc_info)
                e = error.report.E000(
                    msg + _(' See more information in logs.'))
            if isinstance(e, types.GeneratorType):
                errors.extend(self._to_list(e, select, ignore))
            else:
                if ((select and e.code not in select) or
                        (ignore and e.code in ignore)):
                    LOG.debug('Skipped: {code} {message}'
                              ''.format(**e.to_dict()))
                    continue
                LOG.debug('Reported: {code} {message}'
                          ''.format(**e.to_dict()))
                errors.append(e)

        return sorted(errors, key=lambda err: err.code)

    def load_plugins(self):
        if self.plugins is not None:
            return
        self.plugins = stevedore.ExtensionManager(
            'muranopkgcheck.plugins', invoke_on_load=True,
            propagate_map_exceptions=True,
            on_load_failure_callback=self.failure_hook)
        plugin_validators = list(itertools.chain(
            *(p.obj.validators() for p in self.plugins)
        ))
        self.validators += plugin_validators

    @staticmethod
    def failure_hook(_, ep, err):
        LOG.error(_LE('Could not load {plugin}: {error}'
                      '').format(plugin=ep.name, error=err))
        raise err

    def validate(self, validators=None, select=None, ignore=None,
                 only_errors=False):
        validators = validators or self.validators
        report_chains = []
        for validator in validators:
            v = validator(self.pkg)
            report_chains.append(v.run())
        issues = self._to_list(itertools.chain(*report_chains), select, ignore)
        if only_errors:
            return [err for err in issues if err.is_error()]
        else:
            return issues



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import six

from muranopkgcheck.checkers import yaql_checker
from muranopkgcheck import error


ASSIGMENT_KEY = re.compile('^\$.?[\w]')


def check_req(check, required=True):
    return locals()

error.register.E203(description='Value should be string type')
error.register.E200(description='No value should be here')
error.register.E204(description='Wrong code structure/assigment')
error.register.E202(description='Not a valid yaql expression')
error.register.W202(description='Not a valid yaql expression')
error.register.E201(description='Not a valid variable name')


CODE_STRUCTURE = {
    'Try': {
        'keywords': {
            'Try': check_req('codeblock'),
            'Catch': check_req('codeblock'),
            'Else': check_req('codeblock', False),
            'Finally': check_req('codeblock', False)}},
    'As': {
        'keywords': {
            'As': check_req('string'),
            'With': check_req('string', False),
            'Do': check_req('codeblock'),
        }
    },
    'Parallel': {
        'keywords': {
            'Limit': check_req('codeblock', False),
            'Parallel': check_req('codeblock')},
    },
    'Repeat': {
        'keywords': {
            'Repeat': check_req('number'),
            'Do': check_req('codeblock')}},
    'If': {
        'keywords': {
            'If': check_req('predicate'),
            'Then': check_req('codeblock'),
            'Else': check_req('codeblock', False)}
    },
    'Break': {
        'keywords': {
            'Break': check_req('empty')}
    },
    'Return': {
        'Return': check_req('expression'),
    },
    'While': {
        'keywords': {
            'While': check_req('predicate'),
            'Do': check_req('codeblock')}
    },
    'For': {
        'keywords': {
            'For': check_req('string'),
            'In': check_req('expression'),
            'Do': check_req('codeblock')}
    },
    'Match': {
        'keywords': {
            'Match': check_req(('expression', 'codeblock')),
            'Value': check_req('expression'),
            'Default': check_req('codeblock'),
        }
    },
    'Switch': {
        'keywords': {
            'Switch': check_req(('predicate', 'codeblock')),
            'Default': check_req('codeblock')}
    },
    'Throw': {
        'keywords': {
            'Throw': check_req('string'),
            'Message': check_req('string')}
    },
    'Continue': {
        'keywords': {
            'Continue': check_req('empty'),
        }
    },
    'Rethrow': {
        'keywords': {
            'Rethrow': check_req('empty'),
        }
    },
}


class CheckCodeStructure(object):
    def __init__(self):
        self._check_mappings = {
            'codeblock': self.codeblock,
            'predicate': self.yaql,
            'empty': self.empty,
            'expression': self.yaql,
            'string': self.string,
            'number': self.yaql,
        }
        self._yaql_checker = yaql_checker.YaqlChecker()

    def string(self, value):
        if not isinstance(value, six.string_types):
            yield error.report.E203('Value of "{0}" should be a string'
                                    ''.format(value), value)

    def empty(self, value):
        if value:
            yield error.report.E200('Statement should be empty, not a '
                                    '"{0}"'.format(value), value)

    def yaql(self, value):
        if not self._yaql_checker(value):
            if isinstance(value, bool):
                return
            yield error.report.W202('"{0}" is not valid yaql expression'
                                    ''.format(value), value)

    def codeblock(self, codeblocks):
        if isinstance(codeblocks, list):
            for block in codeblocks:
                yield self._single_block(block)
        else:
            yield self._single_block(codeblocks)

    def _check_assigment(self, block):
        key = next(iter(block))
        if not isinstance(key, six.string_types) or\
                not ASSIGMENT_KEY.match(key):
            yield error.report.E201('"{0}" is not valid variable name'
                                    ''.format(key), key)

    def _single_block(self, block):
        if isinstance(block, dict):
            yield self._check_structure(block)
        elif isinstance(block, six.string_types):
            yield self.yaql(block)

    def _run_check(self, check, value):
        yield self._check_mappings[check](value)

    def _check_structure(self, block):
        for key, value in six.iteritems(CODE_STRUCTURE):
            if key in block:
                break
        else:
            if len(block.keys()) == 1:
                yield self._check_assigment(block)
            else:
                yield error.report.E204('Wrong code structure/assigment. '
                                        'Probably a typo', block)
            return

        keywords = value.get('keywords', {})
        kset = set(keywords.keys())
        block_keys_set = set(block.keys())
        for missing in (kset - block_keys_set):
            if keywords[missing]['required']:
                yield error.report.E204('Missing keyword "{0}" for "{1}" '
                                        'code structure'
                                        .format(missing, key), block)
        for unknown in (block_keys_set - kset - {key}):
            yield error.report.E201('Unknown keyword "{0}" in "{1}"'
                                    .format(unknown, key), unknown)
        for ckey, cvalue in six.iteritems(keywords):
            check = cvalue['check']
            data = block.get(ckey)
            if not data:
                continue
            if isinstance(check, tuple):
                for left, right in six.iteritems(data):
                    yield self._run_check(check[0], left)
                    yield self._run_check(check[1], right)
            else:
                yield self._run_check(check, data)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock

from muranopkgcheck.tests import test_validator_helpers as helpers
from muranopkgcheck.validators import manifest
from muranopkgcheck import yaml_loader


class ManifestValidatorTests(helpers.BaseValidatorTestClass):
    def setUp(self):
        super(ManifestValidatorTests, self).setUp()
        self._oe_patcher = mock.patch('os.path.exists')
        self.exists = self._oe_patcher.start()
        self.exists.return_value = [True, True]
        self.loaded_package = mock.MagicMock()
        # self.loaded_package = mock.Mock()
        # self.loaded_package.read.return_value.yaml.return_value = [
        #     {'Type': 'Application'}]
        self.mv = manifest.ManifestValidator(self.loaded_package)

    def test_format_as_number(self):
        self.g = self.mv._valid_format(1.3)

    def test_description(self):
        self.g = self.mv._valid_description(yaml_loader.YamlNull())

    def test_description_string(self):
        self.g = self.mv._valid_description("lalal")

    def test_description_number(self):
        self.g = self.mv._valid_description(1.3)
        self.assertIn('Value is not valid string "1.3"',
                      next(self.g).message)

    def test_wrong_format(self):
        self.g = self.mv._valid_format('0.9')
        self.assertIn('Not supported format version "0.9"',
                      next(self.g).message)

    def test_valid_string(self):
        self.g = self.mv._valid_string([])
        self.assertIn('Value is not a string "[]"',
                      next(self.g).message)

    def test_heat_format(self):
        self.g = self.mv._valid_format('Heat/1.0')
        self.assertIn('Not supported format "Heat/1.0"',
                      next(self.g).message)

    def test_heat_format_invalid_version_also(self):
        self.g = self.mv._valid_format('Heat/0.1.0')
        self.assertIn('Not supported format "Heat/0.1.0"',
                      next(self.g).message)

    def test_unsupported_format(self):
        self.g = self.mv._valid_format('Heat.HOT')
        self.assertIn('Not supported format version "Heat.HOT"',
                      next(self.g).message)

    def test_type(self):
        self.g = self.mv._valid_type('Application')

    def test_wrong_type(self):
        self.g = self.mv._valid_type('Shared Library')
        self.assertIn('Type is invalid "Shared Library"', next(self.g).message)

    def test_incorrect_package_version(self):
        self.g = self.mv._valid_version('a1.3')
        self.assertIn('Version format should be compatible with SemVer '
                      'not "a1.3"', next(self.g).message)

    def test_wrong_require_type(self):
        self.g = self.mv._valid_require([1, 2, 3])
        self.assertIn('Require is not a dict type', next(self.g).message)

    def test_wrong_require_fqn(self):
        self.g = self.mv._valid_require({'io.murano!': '1.3.2'})
        self.assertIn('Require key is not valid FQN "io.murano!"',
                      next(self.g).message)

    def test_require(self):
        self.g = self.mv._valid_require({'aaa.bbb': '>= 1.0.0'})

    def test_not_existing_file(self):
        data = {'org.openstack.Flow': 'FlowClassifier.yaml',
                'org.openstack.Instance': 'Instance.yaml'}
        self.loaded_package.search_for.return_value = ['FlowClassifier.yaml']
        self.g = self.mv._valid_classes(data)
        self.assertIn('File "Instance.yaml" is present in Manifest, '
                      'but not in filesystem', next(self.g).message)

    def test_extra_file_in_directory(self):
        data = {'org.openstack.Instance': 'Instance.yaml'}
        self.loaded_package.search_for.return_value = ['FlowClassifier.yaml',
                                                       'Instance.yaml']
        self.g = self.mv._valid_classes(data)
        self.assertIn('File "FlowClassifier.yaml" is not present in Manifest, '
                      'but it is in filesystem', next(self.g).message)

    def test_classess_list(self):
        data = [{'org.openstack.Instance': 'Instance.yaml'}]
        self.loaded_package.search_for.return_value = ['FlowClassifier.yaml',
                                                       'Instance.yaml']
        self.g = self.mv._valid_classes(data)
        self.assertIn('Classes section should be a dict', next(self.g).message)

    def test_missing_ui_file(self):
        self.loaded_package.exists.return_value = False
        self.g = self.mv._valid_ui('ui.yaml')
        self.assertIn('There is no UI file "ui.yaml"',
                      next(self.g).message)

    def test_missing_logo_file(self):
        self.loaded_package.exists.return_value = False
        self.g = self.mv._valid_logo('logo.png')
        self.assertIn('There is no Logo file "logo.png"',
                      next(self.g).message)

    def test_wrong_logo_type(self):
        self.g = self.mv._valid_logo([1, 2, 3])
        self.assertIn('Logo is not a string', next(self.g).message)

    def test_wrong_ui_type(self):
        self.g = self.mv._valid_ui([1, 2, 3])
        self.assertIn('UI is not a string', next(self.g).message)

    def test_tags(self):
        self.g = self.mv._valid_tags(['whatever'])

    def test_tags_false(self):
        self.g = self.mv._valid_tags('whatever')
        self.assertIn('Tags should be a list', next(self.g).message)

    def test_logo_ui_existance(self):
        self.g = self.mv._valid_logo_ui_existance({'Logo': 0, 'UI': 0})

    def test_logo_ui_existance_false(self):
        self.loaded_package.exists.return_value = True
        self.g = self.mv._valid_logo_ui_existance({})

    def test_fullname_wrong(self):
        self.g = self.mv._valid_fullname('aaa.bbb.ccc#')
        self.assertIn('Invalid FullName "aaa.bbb.ccc#"', next(self.g).message)

    def test_fullname(self):
        self.g = self.mv._valid_fullname('invalid.fullname.!@#@')
        self.assertIn('Invalid FullName "invalid.fullname.!@#@"',
                      next(self.g).message)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six
import yaml

__all__ = ['YamlLoader']


class YamlMetadata(object):
    def __init__(self, mark):
        self.mark = mark

    @property
    def line(self):
        return self.mark.line

    @property
    def column(self):
        return self.mark.column

    def get_snippet(self, indent=4, max_length=75):
        return self.mark.get_snippet(indent, max_length)


class YamlObject(object):
    def __init__(self, value=None):
        self.value = value


class YamlMapping(YamlObject, dict):
    pass


class YamlSequence(YamlObject, list):
    pass


class YamlString(YamlObject, six.text_type):
    pass


class YamlNull(YamlObject):
    def __str__(self):
        return 'null'

    def __bool__(self):
        return False
    __nonzero__ = __bool__


BaseLoader = getattr(yaml, 'CSafeLoader', yaml.SafeLoader)


class YamlLoader(BaseLoader):

    def construct_yaml_seq(self, node):
        data = YamlSequence()
        yield data
        data.extend(self.construct_sequence(node))
        data.__yaml_meta__ = node.start_mark

    def construct_yaml_str(self, node):
        value = super(YamlLoader, self).construct_yaml_str(node)
        value = YamlString(value)
        value.__yaml_meta__ = node.start_mark
        return value

    def construct_yaml_map(self, node):
        data = YamlMapping()
        yield data
        value = self.construct_mapping(node)
        data.update(value)
        data.__yaml_meta__ = node.start_mark

    def construct_yaml_null(self, node):
        value = YamlNull(node)
        value.__yaml_meta__ = node.start_mark
        return value

YamlLoader.add_constructor(
    u'tag:yaml.org,2002:seq',
    YamlLoader.construct_yaml_seq)

YamlLoader.add_constructor(
    u'tag:yaml.org,2002:str',
    YamlLoader.construct_yaml_str)

YamlLoader.add_constructor(
    u'tag:yaml.org,2002:map',
    YamlLoader.construct_yaml_map)

YamlLoader.add_constructor(
    u'tag:yaml.org,2002:null',
    YamlLoader.construct_yaml_null)
